require 'rubygems'
require 'active_resource'

# We need a couple monkey patches to get AR working
# just right with the way we present nested resources
module ActiveResource
  class Base
    
    # Allow passing in a parent object in the initialize function
    # If we do so, guess the prefix options by getting the 
    # ids_for_nested_resource out of the parent object
    def initialize_with_parent(attributes = {}, parent = nil)
      @parent = parent
      if @parent
        attributes = @parent.ids_for_nested_resource.merge(attributes) 
      end
      initialize_without_parent(attributes)
    end
    alias_method_chain :initialize, :parent

    def ids_for_nested_resource
      ids = {}
      if @parent
        ids.merge! @parent.ids_for_nested_resource
      end
      ids.merge("#{self.class.name.underscore}_id" => self.id)
    end

    def load(attributes)
      raise ArgumentError, "expected an attributes Hash, got #{attributes.inspect}" unless attributes.is_a?(Hash)
      
      # Retain @prefix_options that we already have by just merging in the ones we get back new
      # This lets us know the full URL to things across #reload calls where the response
      # doesn't come back with all the correct backreferences
      @prefix_options ||= {}
      new_prefix_options, attributes = split_options(attributes)
      @prefix_options.merge! new_prefix_options
      
      # clear the old attributes
      # in our system we always get back all the XML we need
      # and this prevents re-submitting tricky attributes (like template_ref to merge)
      @attributes.clear
      
      attributes.each do |key, value|
        @attributes[key.to_s] =
          case value
            when Array
              resource = find_or_create_resource_for_collection(key)
              value.map do |attrs| 
								attrs = attrs.attributes.dup if attrs.is_a?(resource)
							  resource.new(attrs, self) 
							end
            when Hash
              resource = find_or_create_resource_for(key)
              resource.new(value, self)
            else
              value.dup rescue value
          end
      end
      self
    end
	end
end


# Base class for resources on VLAB
class SkytapResource < ActiveResource::Base
  self.site = 'http://localhost:3000'
  
  class << self
    def headers_with_basic_auth
      headers_without_basic_auth.merge({ 'Authorization' => "Basic #{Base64.encode64(http_basic_user + ':' + http_basic_password).strip}"})
    end
    alias_method_chain :headers, :basic_auth
    
    attr_writer :http_basic_user
    def http_basic_user
      @http_basic_user || superclass.http_basic_user
    end
    
    attr_writer :http_basic_password
    def http_basic_password
      @http_basic_password || superclass.http_basic_password
    end
    
  end
  
  self.http_basic_user = 'admin'
  self.http_basic_password = 'Cincy#1'
  
  def to_s
    "#<#{self.class.name}:#{element_path}>"
  end
end

module RunstateResource
  
  def run!
    self.runstate = 'running'
    save
  end

  def stop!
    self.runstate = 'stopped'
    save
  end

  def halt!
    self.runstate = 'halted'
    save
  end

  def suspend!
    self.runstate = 'suspended'
    save
  end
  
  DEFAULT_TIMEOUT = 10000
  
  def wait_until_running(timeout=DEFAULT_TIMEOUT)
    wait_on_runstate('running', ['booting','resuming','busy'], timeout)
  end
  
  def wait_until_stopped(timeout=DEFAULT_TIMEOUT)
    wait_on_runstate("stopped", ['powering_off','busy'], timeout)
  end

  def wait_until_suspended(timeout=DEFAULT_TIMEOUT)
    wait_on_runstate("suspended", ['suspending','busy'], timeout)
  end
  
  def wait_until_ready
    wait_on_runstate(['running', 'suspended', 'stopped'])
  end
  
  module ClassMethods
    
    # +ignore_states+ == nil means ignore all other states, [] means nothing else valid
    def wait_on_runstates(entities, desired_states, ignore_states=nil, timeout=DEFAULT_TIMEOUT)
      
      desired_states = [desired_states] unless desired_states.is_a?(Array)
      
      return 0 if entities.collect(&:runstate).all?{|s| desired_states.include?(s)}
      
      start_time = Time.now
      Timeout::timeout(timeout) do
        while true
          entities.collect!(&:reload)
          break if entities.collect(&:runstate).all?{|s|desired_states.include?(s)}
          bad_entity = entities.to_a.find do |e| 
            if desired_states.include?(e.runstate)
              false #this guy is good
            else
              if ignore_states
                # bad if we're not in an ignored state
                !ignore_states.include?(e.runstate)
              else
                false 
              end
            end
          end
          if bad_entity
            raise Exception.new("#{bad_entity} had an unexpected runstate of \"#{bad_entity.runstate}\". " \
                 "Was expecting one of #{(ignore_states+desired_states).inspect}")
          else
            sleep(5)
          end
        end
      end
      Time.now - start_time
    end
  end
  
  def self.included(base)
    base.extend ClassMethods
  end
  
  def wait_on_runstate(desired_state, ignore_states=nil, timeout=10000)
    self.class.wait_on_runstates([self], desired_state, ignore_states, timeout)
  end
end

class Project < SkytapResource; end
class Configuration < SkytapResource; include RunstateResource; end
  
class Vm < SkytapResource
  include RunstateResource
end
class Network < SkytapResource
	self.site = "#{SkytapResource.site}/configurations/:configuration_id"
end
class Interface < SkytapResource
	self.site = "#{SkytapResource.site}/vms/:vm_id"
end
class Ip < SkytapResource; end
class Template < SkytapResource; end
class Note < SkytapResource; end
class Service < SkytapResource
	self.site = "#{SkytapResource.site}/vms/:vm_id/interfaces/:interface_id"
end
